import java.awt.Adjustable;
import java.awt.Color;
import java.awt.EventQueue;
import java.awt.Window;

import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import java.io.File;
import java.text.NumberFormat;

import javax.swing.JButton;
import javax.swing.JFormattedTextField;
import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.SwingConstants;
import javax.swing.table.DefaultTableModel;
import javax.swing.text.NumberFormatter;

import algorithms.FindFuncAlgorithm;
import global.GlobalParams;
import outputs.FitnessPanel;
import outputs.OutputConsole;
import outputs.ResultPanel;

import javax.swing.JTextArea;
import javax.swing.JSeparator;
import javax.swing.JSpinner;
import javax.swing.JComboBox;
import java.awt.event.ActionListener;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.ActionEvent;
import javax.swing.border.BevelBorder;
import javax.swing.event.CaretListener;
import javax.swing.event.CaretEvent;
import javax.swing.SpinnerNumberModel;

public class GE_FindFunc_Visual {

	private JFrame frmFunctionApproximation;
	private JSpinner tfPopSize;
	private JSpinner tfCodonSize;
	private JSpinner tfGenerations;
	private JSpinner tfIndSize;
	private JSpinner tfElitism;
	private JSpinner tfBaseMutation;
	private JSpinner tfHighMutation;
	private JSpinner tfCrossover;
	private JTable table;
	private JComboBox<String> cbGrammar;
	private File[] listOfFiles;
	private DefaultTableModel pointsModel;
	private JPanel pnlResult, pnlFitness;
	private boolean isAlgThreadActive = false;
	private FindFuncAlgorithm alg;
	
	private Thread algThread;
	/**
	 * Launch the application.
	 */
	public static void main(String[] args) {
		EventQueue.invokeLater(new Runnable() {
			public void run() {
				try {
					GE_FindFunc_Visual window = new GE_FindFunc_Visual();
					window.frmFunctionApproximation.setVisible(true);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		});
	}

	/**
	 * Create the application.
	 */
	public GE_FindFunc_Visual() {
		initialize();
	}

	/**
	 * Initialize the contents of the frame.
	 */
	private void initialize() {
		frmFunctionApproximation = new JFrame();
		frmFunctionApproximation.setTitle("Function approximation");
		frmFunctionApproximation.setBounds(100, 100, 900, 700);
		frmFunctionApproximation.setResizable(false);
		frmFunctionApproximation.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frmFunctionApproximation.getContentPane().setLayout(null);

	    NumberFormat doubleFormat = NumberFormat.getInstance();
	    NumberFormatter doubleFormatter = new NumberFormatter(doubleFormat);
	    doubleFormatter.setAllowsInvalid(false);
	    doubleFormatter.setMaximum(1.0);
		
		pnlResult = new ResultPanel();
		pnlResult.setBounds(10, 28, 402, 285);//402, 285
		frmFunctionApproximation.getContentPane().add(pnlResult);
		
		pnlFitness = FitnessPanel.getInstance();
		pnlFitness.setBounds(10, 341, 402, 285);
		frmFunctionApproximation.getContentPane().add(pnlFitness);
		
		JButton btnStart = new JButton("Start");
		btnStart.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				start();
			}
		});
		btnStart.setBounds(10, 637, 89, 23);
		frmFunctionApproximation.getContentPane().add(btnStart);
		
		JButton btnStop = new JButton("Stop");
		btnStop.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				if(isAlgThreadActive) {
					algThread.stop();
					ResultPanel p = (ResultPanel)(pnlResult);
					p.init();
				}
			}
		});
		btnStop.setBounds(109, 637, 89, 23);
		frmFunctionApproximation.getContentPane().add(btnStop);
		
		JLabel lblNewLabel = new JLabel("Population size:");
		lblNewLabel.setBounds(422, 11, 147, 14);
		frmFunctionApproximation.getContentPane().add(lblNewLabel);
		
		tfPopSize = new JSpinner();
		tfPopSize.setModel(new SpinnerNumberModel(new Integer(100), new Integer(1), null, new Integer(100)));
		tfPopSize.setValue(GlobalParams.populationSize);
		tfPopSize.setBounds(579, 8, 86, 20);
		frmFunctionApproximation.getContentPane().add(tfPopSize);
		
		JLabel lblCodonSize = new JLabel("Codon size:");
		lblCodonSize.setBounds(422, 39, 147, 14);
		frmFunctionApproximation.getContentPane().add(lblCodonSize);
		
		tfCodonSize = new JSpinner();
		tfCodonSize.setModel(new SpinnerNumberModel(new Integer(8), new Integer(1), null, new Integer(1)));
		tfCodonSize.setValue(GlobalParams.codonSize);
		tfCodonSize.setBounds(579, 36, 86, 20);
		frmFunctionApproximation.getContentPane().add(tfCodonSize);
		
		JLabel lblNumberOfGenerations = new JLabel("Number of generations:");
		lblNumberOfGenerations.setBounds(422, 67, 147, 14);
		frmFunctionApproximation.getContentPane().add(lblNumberOfGenerations);
		
		tfGenerations = new JSpinner();
		tfGenerations.setModel(new SpinnerNumberModel(new Integer(100), new Integer(1), null, new Integer(10)));
		tfGenerations.setValue(GlobalParams.numberOfGenerations);
		tfGenerations.setBounds(579, 64, 86, 20);
		frmFunctionApproximation.getContentPane().add(tfGenerations);
		
		JLabel lblMaxIndividualSize = new JLabel("Max individual size:");
		lblMaxIndividualSize.setBounds(422, 95, 147, 14);
		frmFunctionApproximation.getContentPane().add(lblMaxIndividualSize);
		
		tfIndSize = new JSpinner();
		tfIndSize.setModel(new SpinnerNumberModel(new Integer(20), new Integer(3), null, new Integer(1)));
		tfIndSize.setValue(GlobalParams.maxIndSize);
		tfIndSize.setBounds(579, 92, 86, 20);
		frmFunctionApproximation.getContentPane().add(tfIndSize);
		
		JLabel lblElitism = new JLabel("Elitism:");
		lblElitism.setBounds(422, 123, 147, 14);
		frmFunctionApproximation.getContentPane().add(lblElitism);
		
		tfElitism = new JSpinner();
		tfElitism.setModel(new SpinnerNumberModel(new Integer(1), new Integer(1), null, new Integer(1)));
		tfElitism.setValue(GlobalParams.elitism);
		tfElitism.setBounds(579, 120, 86, 20);
		frmFunctionApproximation.getContentPane().add(tfElitism);
		
		JLabel lblBaseMutation = new JLabel("Base mutation:");
		lblBaseMutation.setBounds(422, 151, 147, 14);
		frmFunctionApproximation.getContentPane().add(lblBaseMutation);
		
		JLabel lblHighestMutation = new JLabel("Highest mutation:");
		lblHighestMutation.setBounds(422, 179, 147, 14);
		frmFunctionApproximation.getContentPane().add(lblHighestMutation);
		
		tfBaseMutation = new JSpinner();
		tfBaseMutation.setModel(new SpinnerNumberModel(0.01, 0.01, 1.0, 0.01));
		tfBaseMutation.setValue(GlobalParams.baseMutationPropability);
		tfBaseMutation.setBounds(579, 148, 86, 20);
		frmFunctionApproximation.getContentPane().add(tfBaseMutation);
		
		tfHighMutation = new JSpinner();
		tfHighMutation.setModel(new SpinnerNumberModel(0.01, 0.01, 1.0, 0.01));
		tfHighMutation.setValue(GlobalParams.highestMutationPropability);
		tfHighMutation.setBounds(579, 176, 86, 20);
		frmFunctionApproximation.getContentPane().add(tfHighMutation);
		
		pointsModel = new DefaultTableModel(new Object[]{"x", "y"}, 0);
		
		table = new JTable(pointsModel);
		table.setEnabled(false);
		JScrollPane jpane = new JScrollPane(table);
		jpane.setSize(145, 157);
		jpane.setLocation(675, 36);
		frmFunctionApproximation.getContentPane().add(jpane);
		
		JLabel lblPoints = new JLabel("Points:");
		lblPoints.setHorizontalAlignment(SwingConstants.CENTER);
		lblPoints.setBounds(675, 11, 145, 14);
		frmFunctionApproximation.getContentPane().add(lblPoints);
		
		JButton btnPoints = new JButton("Edit points");
		btnPoints.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				new GE_FindFunc_Visual_PointsEdit(pointsModel);
			}
		});
		btnPoints.setBounds(675, 203, 145, 23);
		frmFunctionApproximation.getContentPane().add(btnPoints);
		
		JTextArea txtOutput = OutputConsole.getInstance();
		JScrollPane jpane1 = new JScrollPane(txtOutput);
		jpane1.setBounds(422, 292, 460, 368);
		frmFunctionApproximation.getContentPane().add(jpane1);
		
		JSeparator separator = new JSeparator();
		separator.setBounds(422, 268, 460, 2);
		frmFunctionApproximation.getContentPane().add(separator);
		
		JLabel lblNewLabel_1 = new JLabel("Output:");
		lblNewLabel_1.setBounds(422, 276, 46, 14);
		frmFunctionApproximation.getContentPane().add(lblNewLabel_1);
		
		cbGrammar = new JComboBox<String>();
		cbGrammar.setBounds(579, 237, 241, 20);
		frmFunctionApproximation.getContentPane().add(cbGrammar);
		this.fillComboBox(cbGrammar);
		
		JLabel lblGrammarFile = new JLabel("Grammar file:");
		lblGrammarFile.setBounds(422, 240, 147, 14);
		frmFunctionApproximation.getContentPane().add(lblGrammarFile);
		
		JLabel lblNewLabel_2 = new JLabel("Result:");
		lblNewLabel_2.setBounds(10, 11, 53, 14);
		frmFunctionApproximation.getContentPane().add(lblNewLabel_2);
		
		JLabel lblFitnessProgression = new JLabel("Fitness progression:");
		lblFitnessProgression.setHorizontalAlignment(SwingConstants.CENTER);
		lblFitnessProgression.setBounds(10, 324, 402, 14);
		frmFunctionApproximation.getContentPane().add(lblFitnessProgression);
		
		JButton btnThrd = new JButton("3rd solution");
		btnThrd.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				String s = alg.getBestSolutions()[2];
				((ResultPanel) pnlResult).setFunc(s);
				((ResultPanel) pnlResult).init();
			}
		});
		btnThrd.setBounds(299, 7, 114, 23);
		frmFunctionApproximation.getContentPane().add(btnThrd);
		
		JButton btnScnd = new JButton("2nd solution");
		btnScnd.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				String s = alg.getBestSolutions()[1];
				((ResultPanel) pnlResult).setFunc(s);
				((ResultPanel) pnlResult).init();
			}
		});
		btnScnd.setBounds(186, 7, 114, 23);
		frmFunctionApproximation.getContentPane().add(btnScnd);
		
		JButton btnFirst = new JButton("Best solution");
		btnFirst.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				String s = alg.getBestSolutions()[0];
				((ResultPanel) pnlResult).setFunc(s);
				((ResultPanel) pnlResult).init();
			}
		});
		btnFirst.setBounds(73, 7, 114, 23);
		frmFunctionApproximation.getContentPane().add(btnFirst);
		
		JLabel lblCrossoverPropability = new JLabel("Crossover propability");
		lblCrossoverPropability.setBounds(422, 207, 147, 14);
		frmFunctionApproximation.getContentPane().add(lblCrossoverPropability);
		
		tfCrossover = new JSpinner();
		tfCrossover.setModel(new SpinnerNumberModel(0.8, 0.01, 1.0, 0.1));
		tfCrossover.setBounds(579, 204, 86, 20);
		tfCrossover.setValue(GlobalParams.crossoverPropability);
		frmFunctionApproximation.getContentPane().add(tfCrossover);
	}

	protected void start() {
		//if(running) 	zjistime, zda uz neni spusteno
		DefaultTableModel model = (DefaultTableModel) table.getModel();
		if(!(model.getRowCount() > 0)) {
			table.getTableHeader().setBackground(Color.RED);
			return;
		}
		
		GlobalParams.populationSize = (int) this.tfPopSize.getValue();
		GlobalParams.codonSize = (int) this.tfCodonSize.getValue();
		GlobalParams.numberOfGenerations = (int) this.tfGenerations.getValue();
		GlobalParams.elitism = (int) this.tfElitism.getValue();
		GlobalParams.maxIndSize = (int) this.tfIndSize.getValue();
		GlobalParams.baseMutationPropability = (double) this.tfBaseMutation.getValue();
		GlobalParams.mutationPropability = (double) this.tfBaseMutation.getValue();
		GlobalParams.highestMutationPropability = (double) this.tfHighMutation.getValue();
		GlobalParams.crossoverPropability = (double) this.tfCrossover.getValue();
		GlobalParams.grammarFilePath = "grammar/" + listOfFiles[this.cbGrammar.getSelectedIndex()].getName();
		
		double xS[] = new double[table.getRowCount()];
		double yS[] = new double[table.getRowCount()];
		
		for(int i = 0; i < table.getRowCount(); i++) {
			if(model.getValueAt(i, 0).getClass() == String.class) {
				String s = (String) model.getValueAt(i, 0);
				xS[i] = Double.parseDouble(s);
			} else
				xS[i] = (double) model.getValueAt(i, 0);
			if(model.getValueAt(i, 1).getClass() == String.class) {
				String s = (String) model.getValueAt(i, 1);
				yS[i] = Double.parseDouble(s);
			} else
				yS[i] = (double) model.getValueAt(i, 1);
		}

		alg = new FindFuncAlgorithm();
		alg.setxS(xS);
		alg.setValues(yS);
		
		alg.setPanel((ResultPanel) pnlResult);
		
		algThread = new Thread(){
	        public void run(){
	        	try {
	        		//for(int j = 0; j < 50; j++) {
		        		isAlgThreadActive = true;
						alg.execute();
						isAlgThreadActive = false;
	        		//}
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
	        }
	      };
	      algThread.start();
	  	
	  }
	private void fillComboBox(JComboBox<String> cbGrammar) {
		File folder = new File("grammar/");
		listOfFiles = folder.listFiles();

		for (File file : listOfFiles) {
		    if (file.isFile()) {
		    	int i = file.getName().lastIndexOf('.');
		    	if (i > 0) {
		    	    String extension = file.getName().substring(i+1);
		    	    if(extension.toLowerCase().equals("bnf")) {
		    	    	cbGrammar.addItem(file.getName());
		    	    }
		    	}
		    }
		}
	}
}
