package algorithms;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

import fitness.FindFunc_Fitness;
import fitness.Fitness;
import global.GlobalParams;
import operators.Operators;
import outputs.FitnessPanel;
import outputs.OutputConsole;
import outputs.ResultPanel;
import outputs.ResultWindow;
import population.Individual;
import population.MathFunc_Individual;
import population.Population;
import utils.GrammarReader;
import utils.Utils;

public class FindFuncAlgorithm {
	private Fitness fitness;
	private ResultWindow r; 
	private ResultPanel panel;
	private boolean drawResult = false;
	private boolean nonVisual = false;
	private GrammarReader reader;
	private static double[] xS;
	private static double[] values;
	private OutputConsole outputConsole;
	private FitnessPanel pnlFitness;
	private String[] bestSolutions;
	
	public String[] getBestSolutions() {
		return bestSolutions;
	}

	public void execute() throws Exception {
		//show some results
		if(drawResult && r == null) r = new ResultWindow();
		if(!nonVisual) {
			if(panel == null)
				panel = new ResultPanel();
			panel.setFunc(null);
			outputConsole=OutputConsole.getInstance();
			pnlFitness = FitnessPanel.getInstance();
			pnlFitness.init();
		}
		double bestFitness = -1;
		int bestFitnessRepeating = -1;
		fitness = new FindFunc_Fitness(this);
		long startTime = System.currentTimeMillis();
		List<Individual> Pop0Inds = new ArrayList<>();

		if(!nonVisual) {
			panel.setxS(xS);
			panel.setyS(values);
			panel.init();
		}
		
		reader  = new GrammarReader();
		for(int i = 0; i < GlobalParams.populationSize; i++) {
			String ind = Utils.generateBinaryString(4);
			Pop0Inds.add(new MathFunc_Individual(ind));
		}
		Population p = new Population(Pop0Inds);
		int i;
		//System.setProperty("java.util.Arrays.useLegacyMergeSort", "true");
		for(i = 0; i < GlobalParams.numberOfGenerations; i++) {
			outputConsole.writeNewLine("generation: " + (i+1), false);
			fitness.setFitnessValuetoPopulation(p);
			p.getIndividuals().sort(new Comparator<Individual>() {

				@Override
				public int compare(Individual i1, Individual i2) {
					int comp = 0;
					/*if(!(i1.getFitness() == i2.getFitness()))
						comp = i1.compareTo(i2);*/
					if(i1.getFitness() < i2.getFitness()) comp = -1;
					else if(i1.getFitness() > i2.getFitness()) comp = 1;
					return comp;
				}
				
			});
			
			//cast kodu, ktera se stara o navyseni pravdepodobnosti mutace v pripade uviznuti v lok. optimu
			outputConsole.writeNewLine("Best fitness: " + p.getIndividuals().get(0).getFitness(), false);
			if(bestFitness == -1) bestFitness = p.getIndividuals().get(0).getFitness();
			if(bestFitness == p.getIndividuals().get(0).getFitness())
				bestFitnessRepeating++;
			else if(bestFitness > p.getIndividuals().get(0).getFitness()) {
				if(bestFitnessRepeating >= 30)
					outputConsole.writeNewLine("mutation back to normal", false);
				bestFitnessRepeating = 0;
				bestFitness = p.getIndividuals().get(0).getFitness();
				GlobalParams.mutationPropability = GlobalParams.baseMutationPropability;
			}
			//System.out.println( bestFitnessRepeating);
			if(bestFitnessRepeating == 30) {
				outputConsole.writeNewLine("mutation increased", false);
				GlobalParams.mutationPropability = GlobalParams.highestMutationPropability;
			}
			
			//pokud najdeme kvalitniho jedince, nevytvarime dalsi populaci
			if(p.getIndividuals().get(0).getFitness() < 0.009) break;

			if(!(i == GlobalParams.numberOfGenerations-1)) 
				p = Operators.generateNewGeneration(p);
			outputConsole.writeNewLine("_____________", false);
			if(!nonVisual) {
				if(panel.getxS() != null) panel.setxS(xS);
				if(panel.getxS() != null) panel.setyS(values);
				panel.setFunc(Utils.mapGenometoPhenotype(Utils.convertBinaryCodons(p.getIndividuals().get(0).getGenome(), 
						GlobalParams.codonSize), 5, reader));
				if(bestFitnessRepeating == 0) {
					pnlFitness.setNewY(p.getIndividuals().get(0).getFitness());
					pnlFitness.setNewX((double) i);
					pnlFitness.draw();
				}
			}
		}
		
		if(drawResult && nonVisual) {
			r.setxS(xS);
			r.setyS(values);
			r.setFunc(Utils.mapGenometoPhenotype(Utils.convertBinaryCodons(p.getIndividuals().get(0).getGenome(), 
					GlobalParams.codonSize), 5, reader));
			r.init();
		}
		if(!nonVisual) {
			panel.setFunc(Utils.mapGenometoPhenotype(Utils.convertBinaryCodons(p.getIndividuals().get(0).getGenome(), 
					GlobalParams.codonSize), 5, reader));
			panel.init();
			pnlFitness.setNewY(p.getIndividuals().get(0).getFitness());
			pnlFitness.setNewX((double) i);
			pnlFitness.draw();
		}
		
		//ulozit stringy nejlepsich funkci do pole
		int solutionstoPassCount = 3;
		bestSolutions = new String[solutionstoPassCount];
		String solution = ""; 
		for(int j = 0; j <p.getIndividuals().size(); j++) {
			String s = (Utils.mapGenometoPhenotype(Utils.convertBinaryCodons(p.getIndividuals().get(j).getGenome(), 
					GlobalParams.codonSize), 5, reader));
			if(!solution.equals(s)) {
				bestSolutions[bestSolutions.length - solutionstoPassCount] = s;
				solution = s;
				solutionstoPassCount--;
			}
			if(solutionstoPassCount < 1) break;
		}

		outputConsole.writeNewLine("Best solutions:", false);
		outputConsole.writeNewLine("	Solution: " + 
		Utils.mapGenometoPhenotype(Utils.convertBinaryCodons(p.getIndividuals().get(0).getGenome(), 
				GlobalParams.codonSize), 5, reader) + "		Fitness: " +  Math.round(p.getIndividuals().get(0).getFitness()*1000)/1000.0d, true);
		outputConsole.writeNewLine("	Solution: " +bestSolutions[1], false);
		outputConsole.writeNewLine("	Solution: " +bestSolutions[2], false);

		long endTime = System.currentTimeMillis();
		outputConsole.writeNewLine("Time total: " + (endTime - startTime), false);
	}

	public GrammarReader getReader() {
		return reader;
	}

	public void setReader(GrammarReader reader) {
		this.reader = reader;
	}

	public ResultWindow getR() {
		return r;
	}

	public void setR(ResultWindow r) {
		this.r = r;
	}

	public boolean isDrawResult() {
		return drawResult;
	}

	public void setDrawResult(boolean drawResult) {
		this.drawResult = drawResult;
	}

	public double[] getxS() {
		return xS;
	}

	public void setxS(double[] xS) {
		FindFuncAlgorithm.xS = xS;
	}

	public double[] getValues() {
		return values;
	}

	public void setValues(double[] values) {
		FindFuncAlgorithm.values = values;
	}

	public boolean isNonVisual() {
		return nonVisual;
	}

	public void setNonVisual(boolean nonVisual) {
		this.nonVisual = nonVisual;
	}

	public ResultPanel getPanel() {
		return panel;
	}

	public void setPanel(ResultPanel panel) {
		this.panel = panel;
	}
	
}
