package fitness;

import java.util.List;
import java.util.stream.IntStream;

import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;

import algorithms.FindFuncAlgorithm;
import global.GlobalParams;
import outputs.OutputConsole;
import population.Individual;
import population.Population;
import utils.MultiThreaded;
import utils.Utils;

public class FindFunc_Fitness implements Fitness {
	
	private FindFuncAlgorithm alg;
	private OutputConsole outputConsole;
	
	public FindFunc_Fitness (FindFuncAlgorithm alg) {
		this.alg = alg;
		outputConsole=OutputConsole.getInstance();
		if(alg.getValues() == null) {
			double[] xS  = new double[pointsNumber];
			double[] values = new double[pointsNumber];
			double x = -4;
			for(int i = 0; i < pointsNumber; i++) {
				xS[i] = x;
				//values[i] = x/Math.sin(x) + 2;
				values[i] = (x+1)/Math.cos(x) * Math.sin(x);
				//values[i] = 2 + 5*x - 3*(x*x);// + 2*(x*x*x); 
				x += 0.6;
			}
			alg.setxS(xS);
			alg.setValues(values);
			/*xS = new double[10];
			values = new double[10];
			xS[0] = 1;
			values[0] = 1.581;
			xS[1] = 2;
			values[1] = -2.403;
			xS[2] =3;
			values[2] = -1.01;
			xS[3] = 4;
			values[3] = -1.53;
			xS[4] = 5;
			values[4] = 3.525;
			xS[5] = 6;
			values[5] = 1.041;
			xS[6] = 7;
			values[6] = 1.326;
			xS[7] = 8;
			values[7] = -6.873;
			xS[8] = 9;
			values[8] = -1.098;
			xS[9] = 10;
			values[9] = -1.192;*/
		}
	}
	
	private int pointsNumber = 50;
	
	@Override
	public void setFitnessValue(Individual individual) {
		double fitness = 0;
		//
		String func = Utils.mapGenometoPhenotype(Utils.convertBinaryCodons(individual.getGenome(), 
				GlobalParams.codonSize), 5, alg.getReader());
		//long startTime = System.currentTimeMillis();
		for(int i = 0; i < alg.getxS().length; i++) {
			double y1;
			try {
				y1 = RunnableEval.computeValue(alg.getxS()[i], func);
				if(Double.isInfinite(y1) || Double.isNaN(y1)) y1 = 25;
				fitness += (alg.getValues()[i] - y1)*(alg.getValues()[i] - y1); 
			} catch (ScriptException e) {
				e.printStackTrace();
			}
		}
		/*long endTime = System.currentTimeMillis();
		System.out.println(func);
		System.out.println("Computing: " + (endTime - startTime));
		System.out.println("");*/
		if(fitness > 0.09) {
			String f1 = func.replace("(", "");
			f1 = f1.replace(")", "");
			//fitness += f1.length()*f1.length()*0.01;
			fitness += f1.length()/10;
		}
		fitness=fitness;
		individual.setFitness(fitness);
	}

	@Override
	public void setFitnessValuetoPopulation(Population p) {
		p.setSumFitness(0);
		long startTime = System.currentTimeMillis();
		/*IntStream.range(0, p.getIndividuals().size()).forEach(
				nbr -> {
					setFitnessValue(p.getIndividuals().get(nbr));
					p.setSumFitness(p.getSumFitness() + p.getIndividuals().get(nbr).getFitness());
				}
			);*/
		RunnableEval.p = p;
		RunnableEval.alg = this.alg;
		RunnableEval.run();
		p = RunnableEval.p;
		//System.out.println("Population fitness: " + p.getSumFitness());
		long endTime = System.currentTimeMillis();
		outputConsole.writeNewLine("Pop. fitness: "+p.getSumFitness(), false);
		outputConsole.writeNewLine("Time setting fitness: " + (endTime - startTime), false);
	}

	public static class RunnableEval extends MultiThreaded {
		private static Population p;
		private static ScriptEngine engine;
		private static FindFuncAlgorithm alg;
		
		public static void run() {
			final List<Individual> inds = p.getIndividuals();
			 Task task = new Task() {
	                @Override
	                public void run() {
	                	double sumFitness = 0.0f;
	        			for(Individual i : inds) {
	        				double fitness = 0;
	        				double y1;
	        				String func = Utils.mapGenometoPhenotype(Utils.convertBinaryCodons(i.getGenome(), 
	        						GlobalParams.codonSize), 5, alg.getReader());
	        				for(int j = 0; j < alg.getxS().length; j++) {
	        					try {
	        						y1 = computeValue(alg.getxS()[j], func);
	        						if(Double.isInfinite(y1) || Double.isNaN(y1)) y1 = 100;
	        						double d = (alg.getValues()[j] - y1)*(alg.getValues()[j] - y1);
	        						fitness += d; 
	        					} catch (ScriptException e) {
	        						// TODO Auto-generated catch block
	        						e.printStackTrace();
	        					}
	        				}
	        				if(fitness > 0.01) {
	        					String f1 = func.replace("(", "");
	        					f1 = f1.replace(")", "");
	        					//fitness += f1.length()*f1.length()*0.01;
	        					//fitness += f1.length()*(1/xS.length);
	        					fitness += i.getGenome().length()*0.01;
	        				}
	        				i.setFitness(fitness);
	        				sumFitness += fitness;
	        			}
	        	        p.setSumFitness(sumFitness);
	                }
	            };

	            startTask(task);
	        while(MultiThreaded.processing()) { continue; }
	        p.setIndividuals(inds);
		}
		
		public static double computeValue(double x, String function) throws ScriptException {
			if(engine == null) {
				ScriptEngineManager mgr = new ScriptEngineManager();
				engine = mgr.getEngineByName("JavaScript");
			}
			try {
				return Double.valueOf(engine.eval(replaceX(x, function)).toString());
			} catch (Exception e) {
				return 0;
			}
		}
		
		private static String replaceX (double x, String function) {
			String expression = function;
			String s;
			if(x < 0)
				s = "(" + String.valueOf(x) + ")";
			else s = String.valueOf(x);
			if(function.contains("x"))
				expression = function.replaceAll("x", s);
			return expression;
		}
		
	}
}
