package operators;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import global.GlobalParams;
import population.Individual;
import population.MathFunc_Individual;
import population.Population;
import utils.Utils;

public abstract class Operators {
	
	public static String[] performCrossover(String gen1, String gen2) {
		String[] children = new String[2];
		Random r = new Random();
		
		if(r.nextDouble() <= GlobalParams.mutationPropability) {
			int gen1Size = gen1.length() / GlobalParams.codonSize;
			int gen2Size = gen2.length() / GlobalParams.codonSize;
			int gen1ParsePos, gen2ParsePos;
				gen1ParsePos = r.nextInt(gen1Size - 1) + 1;
				int tmpSizeMin = gen1ParsePos + gen2Size - GlobalParams.maxIndSize;
				int tmpSizeMax = GlobalParams.maxIndSize - gen1Size + gen1ParsePos;
				if(tmpSizeMin < 1)tmpSizeMin = 1;
				if(tmpSizeMax > gen2Size)tmpSizeMax = gen2Size;
				if(tmpSizeMin == tmpSizeMax) gen2ParsePos = tmpSizeMax;
				else gen2ParsePos = r.nextInt(tmpSizeMax - tmpSizeMin) + tmpSizeMin;
				

			String[] g1Parsed = { gen1.substring(0, gen1ParsePos * GlobalParams.codonSize), gen1.substring(gen1ParsePos * GlobalParams.codonSize) };
			String[] g2Parsed = { gen2.substring(0, gen2ParsePos * GlobalParams.codonSize), gen2.substring(gen2ParsePos * GlobalParams.codonSize) };
			children[0] = g1Parsed[0] + g2Parsed[1];
			children[1] = g2Parsed[0] + g1Parsed[1];
		} else {
			children[0] = gen1;
			children[1] = gen2;
		}
		return children;
	}

	public static String performMutation(String gen) {
		Random r = new Random();
		int index = r.nextInt(gen.length());
		//System.out.println(gen);
		if (gen.charAt(index) == '0')
			gen = gen.substring(0, index) + '1' + gen.substring(index + 1);
		else
			gen = gen.substring(0, index) + '0' + gen.substring(index + 1);
		//System.out.println(gen);
		return gen;
	}
	
	public static Population generateNewGeneration(Population p, double mutationPropability, int elitism) throws Exception {
		int i = 0;
		List<Individual> newGeneration = new ArrayList<>();
		while(i < p.getIndividuals().size()) {
			int mutationNumber = (int) (p.getIndividuals().size() * mutationPropability);
			if(i < elitism) 
				newGeneration.add(p.getIndividuals().get(i));
			if(i >= (p.getIndividuals().size() - mutationNumber)) {
				Individual i1 = performRoulette(p, true);
				if(i1 == null) i1 = performTournament(p, true);
				String genome = i1.getGenome();
				Individual ind = new MathFunc_Individual(performMutation(genome));
				newGeneration.add(ind);
			} if(i >= elitism && i < (p.getIndividuals().size() - mutationNumber)) {
				Individual i1 = performRoulette(p, true);
				Individual i2 = performRoulette(p, true);
				if(i1 == null) i1 = performTournament(p, true);
				if(i2 == null) i2 = performTournament(p, true);
				String gen1 = i1.getGenome();
				String gen2 = i2.getGenome();
				String[] genes = performCrossover(gen1, gen2);
				newGeneration.add(new MathFunc_Individual(genes[0]));
				newGeneration.add(new MathFunc_Individual(genes[1]));
				i++;
			}
			i++;
			
		}
		return new Population(newGeneration);
	}
	
	public static Population generateNewGeneration(Population p) throws Exception {
		return generateNewGeneration(p, GlobalParams.mutationPropability, GlobalParams.elitism);
	}
	
	public static Individual performTournament(Population p, boolean minimization) {
		Random r = new Random();
		Individual i1 = p.getIndividuals().get(r.nextInt(p.getIndividuals().size() - 1));
		Individual i2 = p.getIndividuals().get(r.nextInt(p.getIndividuals().size() - 1));
		if(i1 == i2) {
			i2 = p.getIndividuals().get(r.nextInt(p.getIndividuals().size() - 1));
		}
		if(i1.getFitness() > i2.getFitness())
			if(minimization) return i2;
			else return i1;
		else return i2;
	}
	
	public static Individual performRoulette(Population p, boolean minimization) {
		Random r = new Random();
		double rFit, tmpSum = 0.0;
		
		if(minimization) {
			double worstFitness = (p.getIndividuals().get(p.getIndividuals().size()-1).getFitness()
					+ p.getIndividuals().get(0).getFitness());
			double tmpSumFitness = (worstFitness * p.getIndividuals().size()) - p.getSumFitness();

			rFit = r.nextDouble() * tmpSumFitness;
			
			for (Individual i : p.getIndividuals()) {
				tmpSum += (worstFitness - i.getFitness());
				if (rFit <= tmpSum)  {
					return i;
				}
			}
			
			return null;
		} else {
			rFit = r.nextDouble() * p.getSumFitness();
			for (Individual i : p.getIndividuals()) {
				tmpSum += i.getFitness();
				if (rFit <= tmpSum)  {
					return i;
				}
			}
			return null;
		}
	}
}
