package utils;

import java.util.Random;

import global.GlobalParams;

public abstract class Utils {
	public static String generateBinaryString(int minCodons, int maxCodons, int codonSize) {
		Random r = new Random();
		int codons = r.nextInt(maxCodons - minCodons + 1) + minCodons;
		int stringLength = (codons-1) * codonSize;
		String s = "";
		if(r.nextDouble() > 0.4) {
			for(int i = 0; i < (codonSize-1); i++) {
				s = s + String.valueOf(r.nextInt(2));
			}
			s = s + "1";
		}
		else {
			for(int i = 0; i < (codonSize); i++) {
				s = s + String.valueOf(r.nextInt(2));
			}
		}
		boolean addDots = false;
		for(int i = 1; i <= stringLength; i++) {
			s = s + String.valueOf(r.nextInt(2));
			if(i%codonSize == 0 && addDots) s = s + ".";
		}
		return s;
	}
	public static String generateBinaryString(int minCodons) {
		return generateBinaryString(minCodons, GlobalParams.maxIndSize, GlobalParams.codonSize);
	}
	
	public static int[] convertBinaryCodons(String codon) {
		return convertBinaryCodons(codon, GlobalParams.codonSize);
	}
	
	public static int[] convertBinaryCodons(String codon, int codonSize) {
		String[] cTemp = codon.split(String.format("(?<=\\G.{%1$d})", codonSize));
		int codons[] = new int[cTemp.length];
		String c = "";
		for(int i = 0; i < cTemp.length; i++) {
			codons[i] = Integer.parseInt(cTemp[i], 2);
			c = c + " " + codons[i];
		}
		return codons;
	}
	
	private static int count;
	private static int[] codons;
	private static int actualPos;
	private static GrammarReader reader;
	private static String newS;
	
	public static String mapGenometoPhenotype (int[] codonsArray, int maxNumberOfCycles, GrammarReader gr) {
		newS = "";
		actualPos = -1;
		codons = codonsArray;
		count = codons.length * maxNumberOfCycles;
		reader = gr;
		convertCodon(0);
		return newS;
	}
	
	
	
	private static void convertCodon(int index) {
		actualPos++;
		if(actualPos <= count) {
			int pos = actualPos%codons.length;
			int ruleNumber = codons[pos]%reader.getRules().get(index).size();
			String rule = reader.getRules().get(index).get(ruleNumber);
 			int[] check = checkIsExpandable(rule);
			if(check.length >= 3) newS = newS + "(";
			for(int i = 0; i < check.length; i++) {
				int t = check[i];
				if(t == -1) {
					newS = newS + rule;
				}
				else if(!rule.startsWith("<") && (i == 0)) {
					String strs[] = rule.split("<");
					for (String string : strs) {
						if(!string.endsWith(">")) {
							rule = string;
							break;
						}
					}
					newS = newS + rule + "(";
				} if(t != -1)
					convertCodon(t);
			}
			if((check.length >= 3 && !newS.equals("0")) || ((check[0] != -1) && !rule.contains("<"))) 
				newS = newS + ")";
		//if((check.length >= 3 && !newS.equals("0"))) newS = newS + ")";
		} else {
			newS = "0";
		}
	}

	private static int[] checkIsExpandable(String rule) {
		int[] indexes;
		if(rule.contains("><")) {
			rule = rule.replaceAll("><", ">|<");
			String rules[] = rule.split("\\|");
			indexes = new int[rules.length];
			for(int i = 0; i < rules.length; i++) {
				if(reader.getHeaders().contains(rules[i]))
					indexes[i] = reader.getHeaders().indexOf(rules[i]);
			}
		} else if(rule.contains(">,<")) {
			String rules[] = rule.split("\\,");
			for (int j = 0; j < rules.length; j++) {
				String r = rules[j];
				
				if(r.contains("<")) {
					int index = r.indexOf("<");
					if(index > 0)
					r = r.split(String.format("(?<=\\G.{%1$d})", index))[1];
				}
				if(r.contains(">")) {
					int index = r.lastIndexOf(">");
					r = r.split(String.format("(?<=\\G.{%1$d})", (index+1)))[0];
				}
				rules[j] = r;
			}
			indexes = new int[rules.length];
			for(int i = 0; i < rules.length; i++) {
				if(reader.getHeaders().contains(rules[i]))
					indexes[i] = reader.getHeaders().indexOf(rules[i]);
			}
		} else {
			indexes = new int[1];
			//pokud nelze dal rozvinout, vraci na pozici [0] hodnotu -1
			indexes[0] = -1;
			String strs[] = rule.split("<");
			for (String string : strs) {
				if(string.endsWith(">")) {
					rule = "<" + string;
					break;
				}
			}
			if(reader.getHeaders().contains(rule))
				indexes[0] = reader.getHeaders().indexOf(rule); 
		}
		return indexes;
		
	}
}
